@@ -2,7 +2,7 @@
 # @Author: Kelvin
 # @Date:   2020-05-18 00:15:00
 # @Last Modified by:   Kelvin
-# @Last Modified time: 2021-05-20 14:21:45
+# @Last Modified time: 2021-07-31 19:08:07
 
 import seaborn as sns
 import pandas as pd
@@ -21,13 +21,19 @@ from scanpy.plotting import palettes
 from time import sleep
 import matplotlib.pyplot as plt
 from itertools import combinations
-from typing import Union, Sequence, Tuple, Dict
+from typing import Union, Sequence, Tuple, Dict, Optional
 from matplotlib.axes import Axes
 from matplotlib.figure import Figure
 from tqdm import tqdm
 
 
-def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Union[None, str] = None, palette: Union[None, Sequence] = None, figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4), save: Union[None, str] = None) -> ggplot:
+def clone_rarefaction(self: Union[AnnData, Dandelion],
+                      color: str,
+                      clone_key: Optional[str] = None,
+                      palette: Optional[Sequence] = None,
+                      figsize: Tuple[Union[int, float], Union[int,
+                                                              float]] = (6, 4),
+                      save: Optional[str] = None) -> ggplot:
     """
     Plots rarefaction curve for cell numbers vs clone size.
 
@@ -37,13 +43,13 @@ def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Un
         `AnnData` or `Dandelion` object.
     color : str
         Column name to split the calculation of clone numbers for a given number of cells for e.g. sample, patient etc.
-    clone_key : str, optional
+    clone_key : str, Optional
         Column name specifying the clone_id column in metadata/obs.
-    palette : Sequence, optional
+    palette : Sequence, Optional
         Color mapping for unique elements in color. Will try to retrieve from AnnData `.uns` slot if present.
     figsize :  Tuple[Union[int,float], Union[int,float]]
         Size of plot.
-    save : str, optional
+    save : str, Optional
         Save path.
 
     Returns
@@ -62,7 +68,8 @@ def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Un
 
     groups = list(set(metadata[color]))
     metadata = metadata[metadata['contig_QC_pass'].isin([True, 'True'])]
-    metadata[clonekey] = metadata[clonekey].cat.remove_unused_categories()
+    if type(metadata[clonekey]) == 'category':
+        metadata[clonekey] = metadata[clonekey].cat.remove_unused_categories()
     res = {}
     for g in groups:
         _metadata = metadata[metadata[color] == g]
@@ -71,8 +78,9 @@ def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Un
 
     # remove those with no counts
     rowsum = res_.sum(axis=1)
-    print('removing due to zero counts:', ', '.join(
-        [res_.index[i] for i, x in enumerate(res_.sum(axis=1) == 0) if x]))
+    print(
+        'removing due to zero counts:', ', '.join(
+            [res_.index[i] for i, x in enumerate(res_.sum(axis=1) == 0) if x]))
     sleep(0.5)
     res_ = res_[~(res_.sum(axis=1) == 0)]
 
@@ -88,10 +96,12 @@ def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Un
         if (n[-1:] != tot[i]):
             n = np.append(n, tot[i])
         rarecurve[res_.index[i]] = [
-            rarefun(np.array(res_.iloc[i, ]), z) for z in n]
+            rarefun(np.array(res_.iloc[i, ]), z) for z in n
+        ]
     y = pd.DataFrame([rarecurve[c] for c in rarecurve]).T
-    pred = pd.DataFrame([np.append(np.arange(1, s, 10), s)
-                         for s in res_.sum(axis=1)], index=res_.index).T
+    pred = pd.DataFrame(
+        [np.append(np.arange(1, s, 10), s) for s in res_.sum(axis=1)],
+        index=res_.index).T
 
     y = y.melt()
     pred = pred.melt()
@@ -113,22 +123,16 @@ def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Un
                     pal = None
 
             if pal is not None:
-                p = (ggplot(pred, aes(x="value", y="yhat", color="variable"))
-                     + theme_classic()
-                     + xlab('number of cells')
-                     + ylab('number of clones')
-                     + ggtitle('rarefaction curve')
-                     + labs(color=color)
-                     + scale_color_manual(values=(pal))
-                     + geom_line())
+                p = (ggplot(pred, aes(x="value", y="yhat", color="variable")) +
+                     theme_classic() + xlab('number of cells') +
+                     ylab('number of clones') + ggtitle('rarefaction curve') +
+                     labs(color=color) + scale_color_manual(values=(pal)) +
+                     geom_line())
             else:
-                p = (ggplot(pred, aes(x="value", y="yhat", color="variable"))
-                     + theme_classic()
-                     + xlab('number of cells')
-                     + ylab('number of clones')
-                     + ggtitle('rarefaction curve')
-                     + labs(color=color)
-                     + geom_line())
+                p = (ggplot(pred, aes(x="value", y="yhat", color="variable")) +
+                     theme_classic() + xlab('number of cells') +
+                     ylab('number of clones') + ggtitle('rarefaction curve') +
+                     labs(color=color) + geom_line())
         else:
             if len(list(set((pred.variable)))) <= 20:
                 pal = palettes.default_20
@@ -140,35 +144,29 @@ def clone_rarefaction(self: Union[AnnData, Dandelion], color: str, clone_key: Un
                 pal = None
 
             if pal is not None:
-                p = (ggplot(pred, aes(x="value", y="yhat", color="variable"))
-                     + theme_classic()
-                     + xlab('number of cells')
-                     + ylab('number of clones')
-                     + ggtitle('rarefaction curve')
-                     + labs(color=color)
-                     + scale_color_manual(values=(pal))
-                     + geom_line())
+                p = (ggplot(pred, aes(x="value", y="yhat", color="variable")) +
+                     theme_classic() + xlab('number of cells') +
+                     ylab('number of clones') + ggtitle('rarefaction curve') +
+                     labs(color=color) + scale_color_manual(values=(pal)) +
+                     geom_line())
             else:
-                p = (ggplot(pred, aes(x="value", y="yhat", color="variable"))
-                     + theme_classic()
-                     + xlab('number of cells')
-                     + ylab('number of clones')
-                     + ggtitle('rarefaction curve')
-                     + labs(color=color)
-                     + geom_line())
+                p = (ggplot(pred, aes(x="value", y="yhat", color="variable")) +
+                     theme_classic() + xlab('number of cells') +
+                     ylab('number of clones') + ggtitle('rarefaction curve') +
+                     labs(color=color) + geom_line())
     else:
-        p = (ggplot(pred, aes(x="value", y="yhat", color="variable"))
-             + theme_classic()
-             + xlab('number of cells')
-             + ylab('number of clones')
-             + ggtitle('rarefaction curve')
-             + labs(color=color)
-             + geom_line())
+        p = (ggplot(pred, aes(x="value", y="yhat", color="variable")) +
+             theme_classic() + xlab('number of cells') +
+             ylab('number of clones') + ggtitle('rarefaction curve') +
+             labs(color=color) + geom_line())
     if save:
         p.save(filename='figures/rarefaction' + str(save),
-               height=plt.rcParams['figure.figsize'][0], width=plt.rcParams['figure.figsize'][1], units='in', dpi=plt.rcParams["savefig.dpi"])
+               height=plt.rcParams['figure.figsize'][0],
+               width=plt.rcParams['figure.figsize'][1],
+               units='in',
+               dpi=plt.rcParams["savefig.dpi"])
 
-    return(p)
+    return (p)
 
 
 def random_palette(n: int) -> Sequence:
@@ -177,10 +175,13 @@ def random_palette(n: int) -> Sequence:
     # if max_colors_needed1 > len(cols):
     cols2 = list(sns.color_palette('husl', n))
     palette = random.sample(sns.xkcd_palette(cols) + cols2, n)
-    return(palette)
+    return (palette)
 
 
-def clone_network(adata: AnnData, basis: str = 'vdj', edges: bool = True, **kwargs) -> Union[Figure, Axes, None]:
+def clone_network(adata: AnnData,
+                  basis: str = 'vdj',
+                  edges: bool = True,
+                  **kwargs) -> Union[Figure, Axes, None]:
     """
     Using scanpy's plotting module to plot the network. Only thing that is changed is the dfault options: `basis = 'bcr'` and `edges = True`.
 
@@ -198,7 +199,17 @@ def clone_network(adata: AnnData, basis: str = 'vdj', edges: bool = True, **kwar
     embedding(adata, basis=basis, edges=edges, **kwargs)
 
 
-def barplot(self: Union[AnnData, Dandelion], color: str, palette: str = 'Set1', figsize: Tuple[Union[int, float], Union[int, float]] = (12, 4), normalize: bool = True, sort_descending: bool = True, title: Union[None, str] = None, xtick_rotation: Union[None, Union[int, float]] = None, min_clone_size: Union[None, int] = None, clone_key: Union[None, str] = None, **kwargs) -> Tuple[Figure, Axes]:
+def barplot(self: Union[AnnData, Dandelion],
+            color: str,
+            palette: str = 'Set1',
+            figsize: Tuple[Union[int, float], Union[int, float]] = (12, 4),
+            normalize: bool = True,
+            sort_descending: bool = True,
+            title: Optional[str] = None,
+            xtick_rotation: Optional[Union[int, float]] = None,
+            min_clone_size: Optional[int] = None,
+            clone_key: Optional[str] = None,
+            **kwargs) -> Tuple[Figure, Axes]:
     """
     A barplot function to plot usage of V/J genes in the data.
 
@@ -216,13 +227,13 @@ def barplot(self: Union[AnnData, Dandelion], color: str, palette: str = 'Set1',
         if True, will return as proportion out of 1, otherwise False will return counts. Default is True.
     sort_descending : bool
         whether or not to sort the order of the plot. Default is True.
-    title : str, optional
+    title : str, Optional
         title of plot.
-    xtick_rotation : int, float, optional
+    xtick_rotation : int, float, Optional
         rotation of x tick labels.
-    min_clone_size : int, optional
+    min_clone_size : int, Optional
         minimum clone size to keep. Defaults to 1 if left as None.
-    clone_key : str, optional
+    clone_key : str, Optional
         column name for clones. None defaults to 'clone_id'.
     **kwargs
         passed to `sns.barplot`.
@@ -278,7 +289,21 @@ def barplot(self: Union[AnnData, Dandelion], color: str, palette: str = 'Set1',
     return fig, ax
 
 
-def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[None, str], figsize: Tuple[Union[int, float], Union[int, float]] = (12, 4), normalize: bool = False, title: Union[None, str] = None, sort_descending: bool = True, xtick_rotation: Union[None, Union[float, int]] = None, hide_legend: bool = True, legend_options: Tuple[str, Tuple[float, float], int] = None, labels: Union[None, Sequence] = None, min_clone_size: Union[None, int] = None, clone_key: Union[None, str] = None, **kwargs) -> Tuple[Figure, Axes]:
+def stackedbarplot(self: Union[AnnData, Dandelion],
+                   color: str,
+                   groupby: Optional[str],
+                   figsize: Tuple[Union[int, float], Union[int,
+                                                           float]] = (12, 4),
+                   normalize: bool = False,
+                   title: Optional[str] = None,
+                   sort_descending: bool = True,
+                   xtick_rotation: Optional[Union[int, float]] = None,
+                   hide_legend: bool = True,
+                   legend_options: Tuple[str, Tuple[float, float], int] = None,
+                   labels: Optional[Sequence] = None,
+                   min_clone_size: Optional[int] = None,
+                   clone_key: Optional[str] = None,
+                   **kwargs) -> Tuple[Figure, Axes]:
     """
     A stackedbarplot function to plot usage of V/J genes in the data split by groups.
 
@@ -296,19 +321,19 @@ def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[N
         if True, will return as proportion out of 1, otherwise False will return counts. Default is True.
     sort_descending : bool
         whether or not to sort the order of the plot. Default is True.
-    title : str, optional
+    title : str, Optional
         title of plot.
-    xtick_rotation: Union[None, Union[float,int]] : int, float, optional
+    xtick_rotation: Optional[Union[int,float]] : int, float, Optional
         rotation of x tick labels.
     hide_legend : bool
         whether or not to hide the legend.
     legend_options : Tuple[str, Tuple[float, float], int]
         a tuple holding 3 options for specify legend options: 1) loc (string), 2) bbox_to_anchor (tuple), 3) ncol (int).
-    labels : Sequence, optional
+    labels : Sequence, Optional
         Names of objects will be used for the legend if list of multiple dataframes supplied.
-    min_clone_size : int, optional
+    min_clone_size : int, Optional
         minimum clone size to keep. Defaults to 1 if left as None.
-    clone_key : str, optional
+    clone_key : str, Optional
         column name for clones. None defaults to 'clone_id'.
     **kwargs
         other kwargs passed to `matplotlib.plt`.
@@ -339,7 +364,8 @@ def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[N
     data_ = data[data[clone_].isin(keep)]
 
     dat_ = pd.DataFrame(data_.groupby(color)[groupby].value_counts(
-        normalize=normalize).unstack(fill_value=0).stack(), columns=['value'])
+        normalize=normalize).unstack(fill_value=0).stack(),
+        columns=['value'])
     dat_.reset_index(drop=False, inplace=True)
     dat_order = pd.DataFrame(data[color].value_counts(normalize=normalize))
     dat_ = dat_.pivot(index=color, columns=groupby, values='value')
@@ -350,7 +376,17 @@ def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[N
     elif sort_descending is None:
         dat_ = dat_.sort_index()
 
-    def _plot_bar_stacked(dfall: pd.DataFrame, labels: Union[None, Sequence] = None, figsize: Tuple[Union[int, float], Union[int, float]] = (12, 4), title: str = "multiple stacked bar plot", xtick_rotation: Union[None, Union[float, int]] = None, legend_options: Tuple[str, Tuple[float, float], int] = None, hide_legend: bool = True, H: Literal["/"] = "/", **kwargs) -> Tuple[Figure, Axes]:
+    def _plot_bar_stacked(dfall: pd.DataFrame,
+                          labels: Optional[Sequence] = None,
+                          figsize: Tuple[Union[int, float],
+                                         Union[int, float]] = (12, 4),
+                          title: str = "multiple stacked bar plot",
+                          xtick_rotation: Optional[Union[int, float]] = None,
+                          legend_options: Tuple[str, Tuple[float, float],
+                                                int] = None,
+                          hide_legend: bool = True,
+                          H: Literal["/"] = "/",
+                          **kwargs) -> Tuple[Figure, Axes]:
         """
         Given a list of dataframes, with identical columns and index, create a clustered stacked bar plot.
 
@@ -380,12 +416,13 @@ def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[N
                          legend=False,
                          grid=False,
                          **kwargs)  # make bar plots
-        h, l = ax.get_legend_handles_labels()  # get the handles we want to modify
+        h, l = ax.get_legend_handles_labels(
+        )  # get the handles we want to modify
         for i in range(0, n_df * n_col, n_col):  # len(h) = n_col * n_df
             for j, pa in enumerate(h[i:i + n_col]):
                 for rect in pa.patches:  # for each index
-                    rect.set_x(rect.get_x() + 1 / float(n_df + 1)
-                               * i / float(n_col))
+                    rect.set_x(rect.get_x() +
+                               1 / float(n_df + 1) * i / float(n_col))
                     rect.set_hatch(H * int(i / n_col))  # edited part
                     rect.set_width(1 / float(n_df + 1))
         ax.set_xticks((np.arange(0, 2 * n_ind, 2) + 1 / float(n_df + 1)) / 2.)
@@ -404,11 +441,20 @@ def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[N
         else:
             Legend = legend_options
         if hide_legend is False:
-            l1 = ax.legend(h[:n_col], l[:n_col], loc=Legend[0],
-                           bbox_to_anchor=Legend[1], ncol=Legend[2], frameon=False)
+            l1 = ax.legend(h[:n_col],
+                           l[:n_col],
+                           loc=Legend[0],
+                           bbox_to_anchor=Legend[1],
+                           ncol=Legend[2],
+                           frameon=False)
             if labels is not None:
-                l2 = plt.legend(
-                    n, labels, loc=Legend[0], bbox_to_anchor=Legend[1], ncol=Legend[2], frameon=False)
+                l2 = plt.legend(n,
+                                labels,
+                                loc=Legend[0],
+                                bbox_to_anchor=Legend[1],
+                                ncol=Legend[2],
+                                frameon=False)
+                ax.add_artist(l2)
             ax.add_artist(l1)
         if xtick_rotation is None:
             plt.xticks(rotation=90)
@@ -423,10 +469,29 @@ def stackedbarplot(self: Union[AnnData, Dandelion], color: str, groupby: Union[N
     else:
         title = title
 
-    return _plot_bar_stacked(dat_, labels=labels, figsize=figsize, title=title, xtick_rotation=xtick_rotation, legend_options=legend_options, hide_legend=hide_legend, **kwargs)
+    return _plot_bar_stacked(dat_,
+                             labels=labels,
+                             figsize=figsize,
+                             title=title,
+                             xtick_rotation=xtick_rotation,
+                             legend_options=legend_options,
+                             hide_legend=hide_legend,
+                             **kwargs)
 
 
-def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus: str, clone_key: Union[None, str] = None, figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4), width: Union[None, Union[int, float]] = None, title: Union[None, str] = None, xtick_rotation: Union[None, Union[float, int]] = None, hide_legend: bool = True, legend_options: Tuple[str, Tuple[float, float], int] = None, labels: Union[None, Sequence] = None, **kwargs) -> Tuple[Figure, Axes]:
+def spectratype(self: Union[AnnData, Dandelion],
+                color: str,
+                groupby: str,
+                locus: str,
+                clone_key: Optional[str] = None,
+                figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4),
+                width: Optional[Union[int, float]] = None,
+                title: Optional[str] = None,
+                xtick_rotation: Optional[Union[int, float]] = None,
+                hide_legend: bool = True,
+                legend_options: Tuple[str, Tuple[float, float], int] = None,
+                labels: Optional[Sequence] = None,
+                **kwargs) -> Tuple[Figure, Axes]:
     """
     A spectratype function to plot usage of CDR3 length in the data split by groups.
 
@@ -442,24 +507,24 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
         either IGH or IGL.
     figsize : Tuple[Union[int,float], Union[int,float]]
         figure size. Default is (6, 4).
-    width : float, int, optional
+    width : float, int, Optional
         width of bars.
-    title : str, optional
+    title : str, Optional
         title of plot.
-    xtick_rotation : int, float, optional
+    xtick_rotation : int, float, Optional
         rotation of x tick labels.
     hide_legend : bool
         whether or not to hide the legend.
     legend_options : Tuple[str, Tuple[float, float], int]
         a tuple holding 3 options for specify legend options: 1) loc (string), 2) bbox_to_anchor (tuple), 3) ncol (int).
-    labels : Sequence, optional
+    labels : Sequence, Optional
         Names of objects will be used for the legend if list of multiple dataframes supplied.
     **kwargs
         other kwargs passed to matplotlib.pyplot.plot
 
     Returns
     -------
-    sectratype plot
+    spectratype plot
     """
 
     if clone_key is None:
@@ -474,7 +539,8 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
             data = self.copy()
         except:
             AttributeError(
-                "Please provide a <class 'Dandelion'> class object or a pandas dataframe instead of %s." % self.__class__)
+                "Please provide a <class 'Dandelion'> class object or a pandas dataframe instead of %s."
+                % self.__class__)
 
     if 'locus' not in data.columns:
         raise AttributeError("Please ensure dataframe contains 'locus' column")
@@ -484,7 +550,8 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
     data = data[data['locus'].isin(locus)]
     data[groupby] = [str(l) for l in data[groupby]]
     dat_ = pd.DataFrame(data.groupby(color)[groupby].value_counts(
-        normalize=False).unstack(fill_value=0).stack(), columns=['value'])
+        normalize=False).unstack(fill_value=0).stack(),
+        columns=['value'])
     dat_.reset_index(drop=False, inplace=True)
     dat_[color] = pd.to_numeric(dat_[color], errors='coerce')
     dat_.sort_values(by=color)
@@ -492,7 +559,19 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
     new_index = range(0, int(dat_[color].max()) + 1)
     dat_2 = dat_2.reindex(new_index, fill_value=0)
 
-    def _plot_spectra_stacked(dfall: pd.DataFrame, labels: Union[None, Sequence] = None, figsize: Tuple[Union[int, float], Union[int, float]] = (6, 4), title: str = "multiple stacked bar plot", width: Union[None, Union[int, float]] = None, xtick_rotation: Union[None, Union[float, int]] = None, legend_options: Tuple[str, Tuple[float, float], int] = None, hide_legend: bool = True, H: Literal["/"] = "/", **kwargs) -> Tuple[Figure, Axes]:
+    def _plot_spectra_stacked(dfall: pd.DataFrame,
+                              labels: Optional[Sequence] = None,
+                              figsize: Tuple[Union[int, float],
+                                             Union[int, float]] = (6, 4),
+                              title: str = "multiple stacked bar plot",
+                              width: Optional[Union[int, float]] = None,
+                              xtick_rotation: Optional[Union[int,
+                                                             float]] = None,
+                              legend_options: Tuple[str, Tuple[float, float],
+                                                    int] = None,
+                              hide_legend: bool = True,
+                              H: Literal["/"] = "/",
+                              **kwargs) -> Tuple[Figure, Axes]:
         if type(dfall) is not list:
             dfall = [dfall]
         n_df = len(dfall)
@@ -512,19 +591,22 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
                          legend=False,
                          grid=False,
                          **kwargs)  # make bar plots
-        h, l = ax.get_legend_handles_labels()  # get the handles we want to modify
+        h, l = ax.get_legend_handles_labels(
+        )  # get the handles we want to modify
         for i in range(0, n_df * n_col, n_col):  # len(h) = n_col * n_df
             for j, pa in enumerate(h[i:i + n_col]):
                 for rect in pa.patches:  # for each index
-                    rect.set_x(rect.get_x() + 1 / float(n_df + 1)
-                               * i / float(n_col))
+                    rect.set_x(rect.get_x() +
+                               1 / float(n_df + 1) * i / float(n_col))
                     rect.set_hatch(H * int(i / n_col))  # edited part
                     # need to see if there's a better way to toggle this.
                     rect.set_width(wdth)
 
         n = 5  # Keeps every 5th label visible and hides the rest
-        [l.set_visible(False) for (i, l) in enumerate(
-            ax.xaxis.get_ticklabels()) if i % n != 0]
+        [
+            l.set_visible(False)
+            for (i, l) in enumerate(ax.xaxis.get_ticklabels()) if i % n != 0
+        ]
         ax.set_title(title)
         ax.set_ylabel('count')
         # Add invisible data to add another legend
@@ -536,11 +618,19 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
         else:
             Legend = legend_options
         if hide_legend is False:
-            l1 = ax.legend(h[:n_col], l[:n_col], loc=Legend[0],
-                           bbox_to_anchor=Legend[1], ncol=Legend[2], frameon=False)
+            l1 = ax.legend(h[:n_col],
+                           l[:n_col],
+                           loc=Legend[0],
+                           bbox_to_anchor=Legend[1],
+                           ncol=Legend[2],
+                           frameon=False)
             if labels is not None:
-                l2 = plt.legend(
-                    n, labels, loc=Legend[0], bbox_to_anchor=Legend[1], ncol=Legend[2], frameon=False)
+                l2 = plt.legend(n,
+                                labels,
+                                loc=Legend[0],
+                                bbox_to_anchor=Legend[1],
+                                ncol=Legend[2],
+                                frameon=False)
             ax.add_artist(l1)
         if xtick_rotation is None:
             plt.xticks(rotation=0)
@@ -549,10 +639,34 @@ def spectratype(self: Union[AnnData, Dandelion], color: str, groupby: str, locus
 
         return fig, ax
 
-    return _plot_spectra_stacked(dat_2, labels=labels, figsize=figsize, title=title, width=width, xtick_rotation=xtick_rotation, legend_options=legend_options, hide_legend=hide_legend, **kwargs)
+    return _plot_spectra_stacked(dat_2,
+                                 labels=labels,
+                                 figsize=figsize,
+                                 title=title,
+                                 width=width,
+                                 xtick_rotation=xtick_rotation,
+                                 legend_options=legend_options,
+                                 hide_legend=hide_legend,
+                                 **kwargs)
 
 
-def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, min_clone_size: Union[None, int] = None, clone_key: Union[None, str] = None, color_mapping: Union[None, Sequence, Dict] = None, node_labels: bool = True, node_label_layout: Literal[None, 'rotation', 'numbers'] = 'rotation', group_label_position: Literal['beginning', 'middle', 'end'] = 'middle', group_label_offset: int = 8, figsize: Tuple[Union[int, float], Union[int, float]] = (8, 8), return_graph: bool = False, save: Union[None, str] = None, **kwargs):
+def clone_overlap(self: Union[AnnData, Dandelion],
+                  groupby: str,
+                  colorby: str,
+                  min_clone_size: Optional[int] = None,
+                  clone_key: Optional[str] = None,
+                  color_mapping: Optional[Union[Sequence, Dict]] = None,
+                  node_labels: bool = True,
+                  node_label_layout: Literal[None, 'rotation',
+                                             'numbers'] = 'rotation',
+                  group_label_position: Literal['beginning', 'middle',
+                                                'end'] = 'middle',
+                  group_label_offset: int = 8,
+                  figsize: Tuple[Union[int, float], Union[int,
+                                                          float]] = (8, 8),
+                  return_graph: bool = False,
+                  save: Optional[str] = None,
+                  **kwargs):
     """
     A plot function to visualise clonal overlap as a circos-style plot. Requires nxviz.
 
@@ -564,15 +678,15 @@ def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, m
         column name in obs/metadata for collapsing to nodes in circos plot.
     colorby : str
         column name in obs/metadata for grouping and color of nodes in circos plot.
-    min_clone_size : int, optional
+    min_clone_size : int, Optional
         minimum size of clone for plotting connections. Defaults to 2 if left as None.
-    clone_key : str, optional
+    clone_key : str, Optional
         column name for clones. None defaults to 'clone_id'.
-    color_maopping : Dict, Sequence, optional
+    color_maopping : Dict, Sequence, Optional
         custom color mapping provided as a sequence (correpsonding to order of categories or alpha-numeric order if dtype is not category), or dictionary containing custom {category:color} mapping.
-    node_labels : bool, optional
+    node_labels : bool, Optional
         whether to use node objects as labels or not
-    node_label_layout : bool, optional
+    node_label_layout : bool, Optional
         which/whether (a) node layout is used. One of 'rotation', 'numbers' or None.
     group_label_position : str
         The position of the group label. One of 'beginning', 'middle' or 'end'.
@@ -593,7 +707,9 @@ def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, m
     try:
         import nxviz as nxv
     except:
-        raise(ImportError("Unable to import module `nxviz`. Have you done install nxviz? Try pip install git+https://github.com/zktuong/nxviz.git"))
+        raise (ImportError(
+            "Unable to import module `nxviz`. Have you done install nxviz? Try pip install git+https://github.com/zktuong/nxviz.git"
+        ))
 
     if min_clone_size is None:
         min_size = 2
@@ -667,14 +783,21 @@ def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, m
     edges = {}
     for x in overlap.index:
         if overlap.loc[x].sum() > 1:
-            edges[x] = [y + ({str(clone_): x},) for y in list(combinations(
-                [i for i in overlap.loc[x][overlap.loc[x] == 1].index], 2))]
+            edges[x] = [
+                y + ({
+                    str(clone_): x
+                }, ) for y in list(
+                    combinations(
+                        [i for i in overlap.loc[x][overlap.loc[x] == 1].index],
+                        2))
+            ]
 
     # create graph
     G = nx.Graph()
     # add in the nodes
-    G.add_nodes_from([(p, {str(colorby): d})
-                      for p, d in zip(data[groupby], data[colorby])])
+    G.add_nodes_from([(p, {
+        str(colorby): d
+    }) for p, d in zip(data[groupby], data[colorby])])
 
     # unpack the edgelist and add to the graph
     for edge in edges:
@@ -684,8 +807,9 @@ def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, m
         if self.__class__ == AnnData:
             if pd.api.types.is_categorical_dtype(self.obs[groupby]):
                 try:
-                    colorby_dict = dict(zip(
-                        list(self.obs[str(colorby)].cat.categories), self.uns[str(colorby) + '_colors']))
+                    colorby_dict = dict(
+                        zip(list(self.obs[str(colorby)].cat.categories),
+                            self.uns[str(colorby) + '_colors']))
                 except:
                     pass
     else:
@@ -694,7 +818,8 @@ def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, m
         else:
             if pd.api.types.is_categorical_dtype(data[groupby]):
                 colorby_dict = dict(
-                    zip(list(data[str(colorby)].cat.categories), color_mapping))
+                    zip(list(data[str(colorby)].cat.categories),
+                        color_mapping))
             else:
                 colorby_dict = dict(
                     zip(sorted(list(set(data[str(colorby)]))), color_mapping))
@@ -724,4 +849,4 @@ def clone_overlap(self: Union[AnnData, Dandelion], groupby: str, colorby: str, m
     if save is not None:
         plt.savefig(save, bbox_inches='tight', **kwargs)
     if return_graph:
-        return(c)
+        return (c)
