@@ -2,7 +2,7 @@
 # @Author: Kelvin
 # @Date:   2020-08-13 21:08:53
 # @Last Modified by:   Kelvin
-# @Last Modified time: 2021-06-17 15:20:30
+# @Last Modified time: 2021-07-31 22:16:53
 
 import pandas as pd
 import numpy as np
@@ -16,21 +16,18 @@ from anndata import AnnData
 from skbio.diversity.alpha import chao1, gini_index, shannon
 from tqdm import tqdm
 from time import sleep
-try:
-    from scanpy import logging as logg
-except ImportError:
-    pass
-from typing import Union, Sequence, Tuple, Dict
+from scanpy import logging as logg
+from typing import Union, Dict, Optional
 import warnings
 
 
 def clone_rarefaction(
         self: Union[Dandelion, AnnData],
         groupby: str,
-        clone_key: Union[None, str] = None,
-        diversity_key: Union[None, str] = None) -> Union[AnnData, Dict]:
+        clone_key: Optional[str] = None,
+        diversity_key: Optional[str] = None) -> Union[AnnData, Dict]:
     """
-    Returns rarefaction predictions for cell numbers vs clone size.
+    Return rarefaction predictions for cell numbers vs clone size.
 
     Parameters
     ----------
@@ -38,9 +35,9 @@ def clone_rarefaction(
         `Dandelion` or `AnnData` object.
     groupby : str
         Column name to split the calculation of clone numbers for a given number of cells for e.g. sample, patient etc.
-    clone_key : str, optional
+    clone_key : str, Optional
         Column name specifying the clone_id column in metadata/obs.
-    diversity_key : str, optional
+    diversity_key : str, Optional
         key for 'diversity' results in AnnData's `.uns`.
 
     Returns
@@ -51,6 +48,9 @@ def clone_rarefaction(
 
     if self.__class__ == AnnData:
         metadata = self.obs.copy()
+    elif self.__class__ == Dandelion:
+        metadata = self.metadata.copy()
+
     if clone_key is None:
         clonekey = 'clone_id'
     else:
@@ -58,7 +58,8 @@ def clone_rarefaction(
 
     groups = list(set(metadata[groupby]))
     metadata = metadata[metadata['contig_QC_pass'].isin([True, 'True'])]
-    metadata[clonekey] = metadata[clonekey].cat.remove_unused_categories()
+    if type(metadata[clonekey]) == 'category':
+        metadata[clonekey] = metadata[clonekey].cat.remove_unused_categories()
     res = {}
     for g in groups:
         _metadata = metadata[metadata[groupby] == g]
@@ -66,7 +67,6 @@ def clone_rarefaction(
     res_ = pd.DataFrame.from_dict(res, orient='index')
 
     # remove those with no counts
-    rowsum = res_.sum(axis=1)
     print(
         'removing due to zero counts:', ', '.join(
             [res_.index[i] for i, x in enumerate(res_.sum(axis=1) == 0) if x]))
@@ -74,7 +74,7 @@ def clone_rarefaction(
 
     # set up for calculating rarefaction
     tot = res_.apply(sum, axis=1)
-    S = res_.apply(lambda x: x[x > 0].shape[0], axis=1)
+    # S = res_.apply(lambda x: x[x > 0].shape[0], axis=1)
     nr = res_.shape[0]
 
     # append the results to a dictionary
@@ -115,20 +115,21 @@ def clone_diversity(
     self: Union[Dandelion, AnnData],
     groupby: str,
     method: Literal['gini', 'chao1', 'shannon'] = 'gini',
-    metric: Literal['clone_vertexsize', 'clone_degree',
+    metric: Literal['clone_network', 'clone_degree',
                     'clone_centrality'] = None,
-    clone_key: Union[None, str] = None,
+    clone_key: Optional[str] = None,
     update_obs_meta: bool = True,
-    diversity_key: Union[None, str] = None,
+    diversity_key: Optional[str] = None,
     resample: bool = False,
-    downsample: Union[None, int] = None,
+    downsample: Optional[int] = None,
     n_resample: int = 50,
     normalize: bool = True,
     reconstruct_network: bool = True,
     expanded_only: bool = False,
     use_contracted: bool = False,
-    key_added: Union[None, str] = None,
-    locus: Union[None, Literal['ig', 'tr-ab', 'tr-gd']] = None
+    key_added: Optional[str] = None,
+    locus: Optional[Literal['ig', 'tr-ab', 'tr-gd']] = None,
+    **kwargs,
 ) -> Union[pd.DataFrame, Dandelion, AnnData]:
     """
     Compute B cell clones diversity : Gini indices, Chao1 estimates, or Shannon entropy.
@@ -141,38 +142,55 @@ def clone_diversity(
         Column name to calculate the gini indices on, for e.g. sample, patient etc.
     method : str
         Method for diversity estimation. Either one of ['gini', 'chao1', 'shannon'].
-    metric : str, optional
-        Metric to use for calculating Gini indices of clones. Accepts one of ['clone_vertexsize', 'clone_degree', 'clone_centrality']. `None` defaults to 'clone_vertexsize'.
-    clone_key : str, optional
+    metric : str, Optional
+        Metric to use for calculating Gini indices of clones.
+        Accepts one of ['clone_network', 'clone_degree', 'clone_centrality'].
+        `None` defaults to 'clone_network'.
+    clone_key : str, Optional
         Column name specifying the clone_id column in metadata.
     update_obs_meta : bool
-        If True, a `pandas` dataframe is returned. If False, function will try to populate the input object's metadata/obs slot.
-    diversity_key : str, optional
+        If True, a `pandas` dataframe is returned.
+        If False, function will try to populate the input object's metadata/obs slot.
+    diversity_key : str, Optional
         key for 'diversity' results in `.uns`.
-    downsample : int, optional
+    downsample : int, Optional
         number of cells to downsample to. If None, defaults to size of smallest group.
     resample : bool
-        Whether or not to randomly sample cells without replacement to the minimum size of groups for the diversity calculation. Default is False.
+        Whether or not to randomly sample cells without replacement to
+        the minimum size of groups for the diversity calculation. Default is False.
     n_resample : int
         Number of times to perform resampling. Default is 50.
     normalize : bool
-        Whether or not to return normalized Shannon Entropy according to https://math.stackexchange.com/a/945172. Default is True.
+        Whether or not to return normalized Shannon Entropy according to https://math.stackexchange.com/a/945172.
+        Default is True.
     reconstruct_network : bool
-        Whether or not to reconstruct the network for Gini Index based measures. Default is True and will reconstruct for each group specified by groupby option.
+        Whether or not to reconstruct the network for Gini Index based measures.
+        Default is True and will reconstruct for each group specified by groupby option.
     expanded_only : bool
         Whether or not to calculate gini indices using expanded clones only. Default is False i.e. use all cells/clones.
     use_contracted : bool
-        Whether or not to perform the gini calculation after contraction of clone network. Only applies to calculation of clone size gini index. Default is False. This is to try and preserve the single-cell properties of the network.
-    key_added : str, list, optional
+        Whether or not to perform the gini calculation after contraction of clone network.
+        Only applies to calculation of clone size gini index. Default is False.
+        This is to try and preserve the single-cell properties of the network.
+    key_added : str, list, Optional
         column names for output.
     locus : str
         Mode of data. Only used for method = 'gini', Accepts one of 'ig', 'tr-ab' or 'tr-gd'. None defaults to 'ig'.
+    **kwargs
+        passed to dandelion.tl.generate_nework
     Returns
     -------
     `pandas` dataframe, `Dandelion` object with updated `.metadata` slot or `AnnData` object with updated `.obs` slot.
     """
     if downsample is not None:
         resample = True
+
+    if locus is None:
+        if self.__class__ == Dandelion:
+            locus = best_guess_locus(self.data)
+        else:
+            locus = 'ig'
+
     if method == 'gini':
         if update_obs_meta:
             diversity_gini(self,
@@ -188,7 +206,8 @@ def clone_diversity(
                            expanded_only=expanded_only,
                            use_contracted=use_contracted,
                            key_added=key_added,
-                           locus=locus)
+                           locus=locus,
+                           **kwargs)
         else:
             return (diversity_gini(self,
                                    groupby=groupby,
@@ -203,7 +222,8 @@ def clone_diversity(
                                    expanded_only=expanded_only,
                                    use_contracted=use_contracted,
                                    key_added=key_added,
-                                   locus=locus))
+                                   locus=locus,
+                                   **kwargs))
     if method == 'chao1':
         if update_obs_meta:
             diversity_chao1(self,
@@ -268,7 +288,7 @@ def clone_networkstats(self: Dandelion,
                 self.distance[x].toarray()
                 for x in self.distance if type(self.distance[x]) is csr_matrix
             ],
-                axis=0)
+                          axis=0)
             A = csr_matrix(dist)
             G = nx.Graph()
             G.add_weighted_edges_from(
@@ -349,22 +369,21 @@ def clone_networkstats(self: Dandelion,
         raise TypeError('Input object must be of {}'.format(Dandelion))
 
 
-def diversity_gini(
-    self: Union[Dandelion, AnnData],
-    groupby: str,
-    metric: Union[None, str] = None,
-    clone_key: Union[None, str] = None,
-    update_obs_meta: bool = False,
-    diversity_key: Union[None, str] = None,
-    resample: bool = False,
-    n_resample: int = 50,
-    downsample: Union[None, int] = None,
-    reconstruct_network: bool = True,
-    expanded_only: bool = False,
-    use_contracted: bool = False,
-    key_added: Union[None, str] = None,
-    locus: Union[None, Literal['ig', 'tr-ab', 'tr-gd']] = None
-) -> Union[pd.DataFrame, Dandelion]:
+def diversity_gini(self: Union[Dandelion, AnnData],
+                   groupby: str,
+                   metric: Optional[str] = None,
+                   clone_key: Optional[str] = None,
+                   update_obs_meta: bool = False,
+                   diversity_key: Optional[str] = None,
+                   resample: bool = False,
+                   n_resample: int = 50,
+                   downsample: Optional[int] = None,
+                   reconstruct_network: bool = True,
+                   expanded_only: bool = False,
+                   use_contracted: bool = False,
+                   key_added: Optional[str] = None,
+                   locus: Optional[Literal['ig', 'tr-ab', 'tr-gd']] = None,
+                   **kwargs) -> Union[pd.DataFrame, Dandelion]:
     """
     Compute B cell clones Gini indices.
 
@@ -374,51 +393,59 @@ def diversity_gini(
         `Dandelion` or `AnnData` object.
     groupby : str
         Column name to calculate the Gini indices on, for e.g. sample, patient etc.
-    metric : str, optional
-        Metric to use for calculating Gini indices of clones. Accepts 'clone_degree' and 'clone_centrality'. Defaults to 'clone_centrality'.
-    clone_key : str, optional
+    metric : str, Optional
+        Metric to use for calculating Gini indices of clones.
+        Accepts one of ['clone_network', 'clone_degree', 'clone_centrality'].
+        Defaults to 'clone_centrality'.
+    clone_key : str, Optional
         Column name specifying the clone_id column in metadata.
     update_obs_meta : bool
-        If True, a `pandas` dataframe is returned. If False, function will try to populate the input object's metadata/obs slot.
-    diversity_key : str, optional
+        If True, a `pandas` dataframe is returned.
+        If False, function will try to populate the input object's metadata/obs slot.
+    diversity_key : str, Optional
         Key for 'diversity' results in `.uns`.
     resample : bool
-        Whether or not to randomly sample cells without replacement to the minimum size of groups for the diversity calculation. Default is False. Resampling will automatically trigger reconstruction of network.
+        Whether or not to randomly sample cells without replacement to
+        the minimum size of groups for the diversity calculation.
+        Default is False. Resampling will automatically trigger reconstruction of network.
     n_resample : int
         Number of times to perform resampling. Default is 50.
-    downsample : int, optional
+    downsample : int, Optional
         number of cells to downsample to. If None, defaults to size of smallest group.
     reconstruct_network : bool
-        Whether or not to reconstruct the network for Gini Index based measures. Default is True and will reconstruct for each group specified by groupby option.
+        Whether or not to reconstruct the network for Gini Index based measures.
+        Default is True and will reconstruct for each group specified by groupby option.
     expanded_only : bool
         Whether or not to calculate gini indices using expanded clones only. Default is False i.e. use all cells/clones.
     use_contracted : bool
-        Whether or not to perform the gini calculation after contraction of clone network. Only applies to calculation of clone size gini index. Default is False. This is to try and preserve the single-cell properties of the network.
-    key_added : str, list, optional
+        Whether or not to perform the gini calculation after contraction of clone network.
+        Only applies to calculation of clone size gini index. Default is False.
+        This is to try and preserve the single-cell properties of the network.
+    key_added : str, list, Optional
         column names for output.
     locus : str
         Mode of data. Accepts one of 'ig', 'tr-ab' or 'tr-gd'. None defaults to 'ig'.
-
+    **kwargs
+        passed to dandelion.tl.generate_nework
     Returns
     -------
     `pandas` dataframe or `Dandelion` object with updated `.metadata` slot.
     """
     start = logg.info('Calculating Gini indices')
 
-    def gini_indices(
-        self: Dandelion,
-        groupby: str,
-        metric: Union[None, str] = None,
-        clone_key: Union[None, str] = None,
-        resample: bool = False,
-        n_resample: int = 50,
-        downsample: Union[None, int] = None,
-        reconstruct_network: bool = True,
-        expanded_only: bool = False,
-        contracted: bool = False,
-        key_added: Union[None, str] = None,
-        locus: Union[None, Literal['ig', 'tr-ab', 'tr-gd']] = None
-    ) -> pd.DataFrame:
+    def gini_indices(self: Dandelion,
+                     groupby: str,
+                     metric: Optional[str] = None,
+                     clone_key: Optional[str] = None,
+                     resample: bool = False,
+                     n_resample: int = 50,
+                     downsample: Optional[int] = None,
+                     reconstruct_network: bool = True,
+                     expanded_only: bool = False,
+                     contracted: bool = False,
+                     key_added: Optional[str] = None,
+                     locus: Optional[Literal['ig', 'tr-ab', 'tr-gd']] = None,
+                     **kwargs) -> pd.DataFrame:
         if self.__class__ == AnnData:
             raise TypeError('Only Dandelion class object accepted.')
         elif self.__class__ == Dandelion:
@@ -491,8 +518,6 @@ def diversity_gini(
                     g_c_v_res)
                 self.metadata['clone_network_cluster_size_gini'] = pd.Series(
                     g_c_c_res)
-        elif met == 'clone_size':
-            print("Computing gini indices for clone size using metadata.")
         elif met == 'clone_centrality':
             print(
                 "Computing gini indices for clone size using metadata and node closeness centrality using network."
@@ -527,7 +552,8 @@ def diversity_gini(
                                                      clone_key=clonekey,
                                                      downsample=minsize,
                                                      verbose=False,
-                                                     locus=locus)
+                                                     locus=locus,
+                                                     **kwargs)
                         if met == 'clone_network':
                             n_n, v_s, c_s = clone_networkstats(
                                 resampled,
@@ -567,9 +593,11 @@ def diversity_gini(
                         elif met == 'clone_degree':
                             clone_degree(resampled, verbose=False)
                         else:
-                            raise ValueError(
-                                'Unknown metric for calculating network stats. Please specify one of `clone_centrality` or `clone_degree`.'
-                            )
+                            raise ValueError((
+                                "Unknown metric for calculating network stats. Please specify "
+                                +
+                                "one of `clone_network`, `clone_centrality` or `clone_degree`."
+                            ))
                         # clone size gini
                         _dat = resampled.metadata.copy()
                         _tab = _dat[clonekey].value_counts()
@@ -581,22 +609,6 @@ def diversity_gini(
                         if met == 'clone_network':
                             sizelist.append(_dat[met +
                                                  '_cluster_size_gini'].mean())
-                        elif met == 'clone_size':
-                            clonesizecounts = np.array(_tab)
-                            clonesizecounts = clonesizecounts[
-                                clonesizecounts > 0]
-                            if len(clonesizecounts) > 1:
-                                # append a single zero for lorenz curve calculation
-                                clonesizecounts = np.append(clonesizecounts, 0)
-                            if len(clonesizecounts) > 0:
-                                g_c = gini_index(clonesizecounts,
-                                                 method='trapezoids')
-                                # probably not needed anymore but keep just in case
-                                if g_c < 0 or np.isnan(g_c):
-                                    g_c = 0
-                            else:
-                                g_c = 0
-                            sizelist.append(g_c)
                         else:
                             clonesizecounts = np.array(_tab)
                             clonesizecounts = clonesizecounts[
@@ -673,7 +685,8 @@ def diversity_gini(
                             generate_network(ddl_dat,
                                              clone_key=clonekey,
                                              verbose=False,
-                                             locus=locus)
+                                             locus=locus,
+                                             **kwargs)
                             n_n, v_s, c_s = clone_networkstats(
                                 ddl_dat,
                                 expanded_only=expanded_only,
@@ -781,7 +794,8 @@ def diversity_gini(
                        expanded_only=expanded_only,
                        contracted=use_contracted,
                        key_added=key_added,
-                       locus=locus)
+                       locus=locus,
+                       **kwargs)
 
     if diversity_key is None:
         diversitykey = 'diversity'
@@ -806,14 +820,14 @@ def diversity_gini(
 def diversity_chao1(
     self: Union[Dandelion, AnnData],
     groupby: str,
-    clone_key: Union[None, str] = None,
+    clone_key: Optional[str] = None,
     update_obs_meta: bool = False,
-    diversity_key: Union[None, str] = None,
+    diversity_key: Optional[str] = None,
     resample: bool = False,
     n_resample: int = 50,
-    downsample: Union[None, int] = None,
-    key_added: Union[None,
-                     str] = None) -> Union[pd.DataFrame, Dandelion, AnnData]:
+    downsample: Optional[int] = None,
+    key_added: Optional[str] = None
+) -> Union[pd.DataFrame, Dandelion, AnnData]:
     """
     Compute B cell clones Chao1 estimates.
 
@@ -823,19 +837,19 @@ def diversity_chao1(
         `Dandelion` or `AnnData` object.
     groupby : str
         Column name to calculate the Chao1 estimates on, for e.g. sample, patient etc.
-    clone_key : str, optional
+    clone_key : str, Optional
         Column name specifying the clone_id column in metadata.
     update_obs_meta : bool
         If True, a `pandas` dataframe is returned. If False, function will try to populate the input object's metadata/obs slot.
-    diversity_key : str, optional
+    diversity_key : str, Optional
         key for 'diversity' results in `.uns`.
     resample : bool
         Whether or not to randomly sample cells without replacement to the minimum size of groups for the diversity calculation. Default is False.
     n_resample : int
         Number of times to perform resampling. Default is 50.
-    downsample : int, optional
+    downsample : int, Optional
         number of cells to downsample to. If None, defaults to size of smallest group.
-    key_added : str, list, optional
+    key_added : str, list, Optional
         column names for output.
 
     Returns
@@ -846,11 +860,11 @@ def diversity_chao1(
 
     def chao1_estimates(self: Union[Dandelion, AnnData],
                         groupby: str,
-                        clone_key: Union[None, str] = None,
+                        clone_key: Optional[str] = None,
                         resample: bool = False,
                         n_resample: int = 50,
-                        downsample: Union[None, int] = None,
-                        key_added: Union[None, str] = None) -> pd.DataFrame:
+                        downsample: Optional[int] = None,
+                        key_added: Optional[str] = None) -> pd.DataFrame:
         if self.__class__ == AnnData:
             metadata = self.obs.copy()
         elif self.__class__ == Dandelion:
@@ -1002,15 +1016,15 @@ def diversity_chao1(
 def diversity_shannon(
     self: Union[Dandelion, AnnData],
     groupby: str,
-    clone_key: Union[None, str] = None,
+    clone_key: Optional[str] = None,
     update_obs_meta: bool = False,
-    diversity_key: Union[None, str] = None,
+    diversity_key: Optional[str] = None,
     resample: bool = False,
     n_resample: int = 50,
     normalize: bool = True,
-    downsample: Union[None, int] = None,
-    key_added: Union[None,
-                     str] = None) -> Union[pd.DataFrame, Dandelion, AnnData]:
+    downsample: Optional[int] = None,
+    key_added: Optional[str] = None
+) -> Union[pd.DataFrame, Dandelion, AnnData]:
     """
     Compute B cell clones Shannon entropy.
 
@@ -1020,11 +1034,11 @@ def diversity_shannon(
         `Dandelion` or `AnnData` object.
     groupby : str
         Column name to calculate the Shannon entropy on, for e.g. sample, patient etc.
-    clone_key : str, optional
+    clone_key : str, Optional
         Column name specifying the clone_id column in metadata.
     update_obs_meta : bool
         If True, a `pandas` dataframe is returned. If False, function will try to populate the input object's metadata/obs slot.
-    diversity_key : str, optional
+    diversity_key : str, Optional
         key for 'diversity' results in `.uns`.
     resample : bool
         Whether or not to randomly sample cells without replacement to the minimum size of groups for the diversity calculation. Default is False.
@@ -1032,9 +1046,9 @@ def diversity_shannon(
         Number of times to perform resampling. Default is 50.
     normalize : bool
         Whether or not to return normalized Shannon Entropy according to https://math.stackexchange.com/a/945172. Default is True.
-    downsample : int, optional
+    downsample : int, Optional
         number of cells to downsample to. If None, defaults to size of smallest group.
-    key_added : str, list, optional
+    key_added : str, list, Optional
         column names for output.
 
     Returns
@@ -1045,12 +1059,12 @@ def diversity_shannon(
 
     def shannon_entropy(self: Union[Dandelion, AnnData],
                         groupby: str,
-                        clone_key: Union[None, str] = None,
+                        clone_key: Optional[str] = None,
                         resample: bool = False,
                         n_resample: int = 50,
                         normalize: bool = True,
-                        downsample: Union[None, int] = None,
-                        key_added: Union[None, str] = None) -> pd.DataFrame:
+                        downsample: Optional[int] = None,
+                        key_added: Optional[str] = None) -> pd.DataFrame:
         if self.__class__ == AnnData:
             metadata = self.obs.copy()
         elif self.__class__ == Dandelion:
